import logging
import sys
from datetime import date, datetime
from logging.handlers import RotatingFileHandler

import sml2mqtt
import sml2mqtt._args


class MidnightRotatingFileHandler(RotatingFileHandler):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.last_check: date = datetime.now().date()

    def shouldRollover(self, record):
        date = datetime.now().date()
        if date == self.last_check:
            return 0
        self.last_check = date
        return super().shouldRollover(record)


def setup_log():
    lvl = sml2mqtt.CONFIG.log.level

    # This is the longest logger name str
    chars = 0
    for device in sml2mqtt.CONFIG.devices:
        chars = max(len(f'sml.device.{device.device}'), chars)
    log_format = logging.Formatter("[{asctime:s}] [{name:" + str(chars) + "s}] {levelname:8s} | {message:s}", style='{')

    handler = MidnightRotatingFileHandler(
        str(sml2mqtt.CONFIG.log.file), maxBytes=1024 * 1024, backupCount=3, encoding='utf-8'
    )
    handler.setFormatter(log_format)
    handler.setLevel(logging.DEBUG)

    # Bind handler to logger
    root = logging.getLogger()
    root.setLevel(lvl)
    root.addHandler(handler)

    if sml2mqtt._args.ARGS.analyze:
        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(lvl)
        ch.setFormatter(log_format)
        root.addHandler(ch)



import asyncio
import typing
import logging

from serial_asyncio import SerialTransport, create_serial_connection
from .sml_device_status import DeviceStatus

if typing.TYPE_CHECKING:
    from .sml_device import Device


log = logging.getLogger('sml.serial')


class SmlSerial(asyncio.Protocol):
    @classmethod
    async def create(cls, url: str, device: 'Device', timeout: float) -> 'SmlSerial':
        transport, protocol = await create_serial_connection(
            asyncio.get_event_loop(), cls, url, baudrate=9600)  # type: SerialTransport, SmlSerial

        protocol.url = url
        protocol.device = device
        protocol.timeout_secs = timeout
        return protocol

    def __init__(self) -> None:
        super().__init__()

        self.url: str = None
        self.device: 'Device' = None

        self.transport: typing.Optional[SerialTransport] = None

        self.timeout_secs: typing.Union[int, float] = 3
        self.timeout_event = asyncio.Event()
        self.timeout_task: typing.Optional[asyncio.Task] = None

        self.pause_task: typing.Optional[asyncio.Task] = None

        self.loop = asyncio.get_event_loop()

    def connection_made(self, transport):
        self.transport = transport
        self.timeout_task = asyncio.create_task(self.watchdog())

        log.debug(f'Port {self.url} successfully opened')

        self.device.set_status(DeviceStatus.PORT_OPENED)

    def data_received(self, data: bytes):
        self.device.stream.add(data)
        self.timeout_event.set()

        self.pause_serial()
        asyncio.ensure_future(self.device.read(), loop=self.loop)

    async def resume_serial(self):
        await asyncio.sleep(0.4, loop=self.loop)

        self.pause_task = None
        self.transport.resume_reading()

    def pause_serial(self):
        self.transport.pause_reading()
        self.pause_task = asyncio.ensure_future(self.resume_serial(), loop=self.loop)

    def connection_lost(self, exc):
        self.close()

        log.info(f'Port {self.url} was closed')
        self.device.set_status(DeviceStatus.PORT_CLOSED)

    def close(self):
        if self.timeout_task is not None:
            self.timeout_task.cancel()
            self.timeout_task = None

        if self.pause_task is not None:
            self.pause_task.cancel()
            self.pause_task = None

        if not self.transport.is_closing():
            self.transport.close()

        self.device.stream.clear()

    async def watchdog(self):
        await asyncio.sleep(0.1)

        while True:
            self.timeout_event.clear()

            try:
                await asyncio.wait_for(self.timeout_event.wait(), self.timeout_secs)
                timeout = False
            except asyncio.TimeoutError:
                timeout = True

            if timeout:
                if self.device.status != DeviceStatus.MSG_TIMEOUT:
                    self.device.stream.clear()
                    log.warning(f'Timeout reading from {self.url}')
                    self.device.set_status(DeviceStatus.MSG_TIMEOUT)



import logging
import traceback
from asyncio import create_task
from time import monotonic
from typing import Any, Dict, List, Optional, Set, Tuple

from smllib import CrcError, SmlStreamReader

import sml2mqtt._args
import sml2mqtt.sml_serial
from ._signals import stop_loop
from .config import CONFIG
from .mqtt import publish
from .sml_device_status import DeviceStatus
from .sml_serial import log as serial_log

ALL: Dict[str, 'Device'] = {}


class Device:
    @classmethod
    async def create(cls, url: str, timeout: float, skip_values: List[str]):
        c = cls(url, set(skip_values))
        c.serial = await sml2mqtt.sml_serial.SmlSerial.create(url, c, timeout)
        ALL[url] = c
        return c

    def __init__(self, url: str, skip_values: Set[str]):
        self.stream = SmlStreamReader()
        self.status: Optional[DeviceStatus] = None

        self.serial: 'sml2mqtt.sml_serial.SmlSerial' = None
        self.device_name = url.lower()

        self.log = logging.getLogger(f'sml.device.{url.split("/")[-1]}')

        self.skip_values = skip_values

        self.value_cache: Dict[str, Tuple[float, Any]] = {}

    def set_status(self, new_status: DeviceStatus):
        if self.status == new_status:
            return None
        self.status = new_status

        # DeviceStatus.PORT_OPENED -> PORT_OPENED
        _stat = str(new_status).split('.')[-1]
        self.log.info(f'{_stat}')

        # Don't publish the port open because we don't have the correct name yet
        if new_status != DeviceStatus.PORT_OPENED:
            create_task(
                publish(CONFIG.mqtt.topics.get_topic(self.device_name, 'status'), _stat)
            )

        # If all ports are closed or we have errors we shut down
        if all(map(lambda x: x.status in (DeviceStatus.PORT_CLOSED, DeviceStatus.ERROR, DeviceStatus.SHUTDOWN),
                   ALL.values())):
            create_task(stop_loop())

    def publish_value(self, name: str, value):
        create_task(
            publish(CONFIG.mqtt.topics.get_topic(self.device_name, name), value)
        )

    async def read(self):
        try:
            try:
                frame = self.stream.get_frame()
                if frame is None:
                    return None
            except CrcError as e:
                serial_log.error(f'Crc error on {self.serial.url}: {e.crc_calc} != {e.crc_msg}')
                self.set_status(DeviceStatus.CRC_ERROR)
                return None

            if sml2mqtt._args.ARGS.analyze:
                self.log.info('')
                self.log.info('Received Frame')
                self.log.info('')
                for obj in frame.parse_frame():
                    for line in obj.format_msg().splitlines():
                        self.log.info(line)
                self.set_status(DeviceStatus.SHUTDOWN)
                return None

            vals = {}
            for sml_obj in frame.get_obis():
                name = sml_obj.obis
                if name in self.skip_values:
                    continue

                value = sml_obj.get_value()

                # We overwrite the device name with the unique identifier
                if name == '0100000009ff':
                    self.device_name = str(value)
                    continue

                # Wh -> kWh
                if sml_obj.unit == 30:
                    # Don't publish disabled energy meters
                    if value < 0.1:
                        continue
                    value = round(value / 1000, 2)

                # Check value cache
                now = monotonic()
                try:
                    c_time, c_value = self.value_cache[name]
                    if value == c_value and now - c_time < CONFIG.general.max_wait:
                        continue
                except KeyError:
                    pass
                self.value_cache[name] = now, value

                # Mark for publishing
                vals[name] = value

            # publish parsed objs
            for n, v in vals.items():
                self.publish_value(n, v)

            # Everything is okay when there are no errors
            if self.status != DeviceStatus.OK:
                self.set_status(DeviceStatus.OK)

        except Exception:
            for line in traceback.format_exc().splitlines():
                self.log.error(line)

            # Signal that an error occured
            self.set_status(DeviceStatus.ERROR)

            # Stop reading from the serial port
            self.serial.close()

        return None



import argparse
import os
import sys
import typing
from pathlib import Path


def find_config_folder(arg_config_path: typing.Optional[str]) -> Path:

    if arg_config_path is None:
        # Nothing is specified, we try to find the config automatically
        check_path = []
        try:
            working_dir = Path(os.getcwd())
            check_path.append(working_dir / 'sml2mqtt')
            check_path.append(working_dir.with_name('sml2mqtt'))
            check_path.append(working_dir.parent.with_name('sml2mqtt'))
        except ValueError:
            # the ValueError gets raised if the working_dir or its parent is empty (e.g. c:\ or /)
            pass

        check_path.append(Path.home() / 'sml2mqtt')   # User home

        # if we run in a venv check the venv, too
        v_env = os.environ.get('VIRTUAL_ENV', '')
        if v_env:
            check_path.append(Path(v_env) / 'sml2mqtt')  # Virtual env dir
    else:
        arg_config_path = Path(arg_config_path)

        # in case the user specifies the config.yml we automatically switch to the parent folder
        if arg_config_path.name.lower() == 'config.yml' and arg_config_path.is_file():
            arg_config_path = arg_config_path.parent

        # Override automatic config detection if something is specified through command line
        check_path = [arg_config_path]

    for config_folder in check_path:
        config_folder = config_folder.resolve()
        if not config_folder.is_dir():
            continue

        config_file = config_folder / 'config.yml'
        if config_file.is_file():
            return config_folder / 'config.yml'

    # we have specified a folder, but the config does not exist so we will create it
    if arg_config_path is not None and arg_config_path.is_dir():
        return arg_config_path / 'config.yml'

    # we have nothing found and nothing specified -> exit
    print('Config file "config.yml" not found!')
    print('Checked folders:\n - ' + '\n - '.join(str(k) for k in check_path))
    print('Please create file or specify a folder with the "-c" arg switch.')
    sys.exit(1)


class CommandArgs(typing.NamedTuple):
    config: typing.Optional[Path]
    analyze: bool


ARGS: CommandArgs


def get_command_line_args(args=None) -> CommandArgs:
    global ARGS

    parser = argparse.ArgumentParser(description='SML to MQTT bridge')
    parser.add_argument(
        '-c',
        '--config',
        help='Path to configuration folder',
        default=None
    )
    parser.add_argument(
        '-a',
        '--analyze',
        help='Processes exactly one sml message and explicitly shows the structure',
        action='store_true',
        default=None
    )
    args = parser.parse_args(args)
    ARGS = CommandArgs(config=find_config_folder(args.config), analyze=args.analyze)
    return ARGS



import asyncio
import logging
import platform
import sys
import traceback
import typing

from sml2mqtt._args import get_command_line_args
from sml2mqtt._log import setup_log
from sml2mqtt._signals import add_shutdown_handler, get_ret_code, shutdown_with_exception
from sml2mqtt.config import CONFIG
from sml2mqtt.mqtt import connect
from sml2mqtt.sml_device import Device


async def a_main():
    await connect()
    await asyncio.sleep(0.1)  # Wait till mqtt is connected

    # Create devices for port
    try:
        for k in CONFIG.devices:
            await Device.create(k.device, k.timeout, k.skip)
    except Exception as e:
        shutdown_with_exception(e)


def main() -> typing.Union[int, str]:
    args = get_command_line_args()
    log = logging.getLogger('sml2mqtt')

    CONFIG.load(args.config)
    setup_log()

    # This is needed to make asyncmqtt work
    if platform.system() == 'Windows':
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

    # Add possibility to stop program with Ctrl + c
    add_shutdown_handler()

    loop = asyncio.get_event_loop()
    try:
        loop.create_task(a_main())
        loop.run_forever()
    except Exception as e:
        for line in traceback.format_exc().splitlines():
            log.error(line)
            print(e)
        return str(e)
    finally:
        loop.run_until_complete(loop.shutdown_asyncgens())
        loop.close()

    return get_ret_code()


if __name__ == "__main__":
    sys.exit(main())



import enum


class DeviceStatus(enum.Enum):
    STARTUP = enum.auto()
    SHUTDOWN = enum.auto()
    PORT_OPENED = enum.auto()
    PORT_CLOSED = enum.auto()
    MSG_TIMEOUT = enum.auto()
    CRC_ERROR = enum.auto()
    ERROR = enum.auto()
    OK = enum.auto()



import logging
import signal
from asyncio import create_task, get_event_loop, sleep
from typing import Optional, Union

log = logging.getLogger('sml2mqtt')
error_text: Optional[str] = None


async def stop_loop():
    await sleep(0.1)
    get_event_loop().stop()


def shutdown_with_exception(e: Exception):
    global error_text
    error_text = str(e)

    log.error(error_text)
    create_task(stop_loop())


def get_ret_code() -> Union[int, str]:
    if error_text is not None:
        return error_text
    return 0


def shutdown_handler(sig, frame):
    print('Shutting down ...')
    log.info('Shutting down ...')
    create_task(stop_loop())


def add_shutdown_handler():
    signal.signal(signal.SIGINT, shutdown_handler)
    signal.signal(signal.SIGTERM, shutdown_handler)



import logging
from pathlib import Path
from typing import Dict, List

from EasyCo import ConfigContainer, ConfigEntry, ConfigFile
from voluptuous import Invalid, Range


class Connection(ConfigContainer):
    client_id: str = 'sml2mqtt'
    host: str = 'localhost'
    port: int = 1883
    user: str = ''
    password: str = ''
    tls: bool = False
    tls_insecure: bool = False


class Topics(ConfigContainer):
    base_topic: str = ConfigEntry('sml2mqtt', key_name='base topic', description='Topic that will prefix all topics')
    last_will: str = ConfigEntry('status', key_name='last will', description='Last will topic')
    alias: Dict[str, str] = ConfigEntry(default_factory=lambda: {'0100010800ff': 'total_energy'}, validator={str: str},
                                        description='These aliases are replaced in the mqtt topics')

    def on_all_values_set(self):
        self.alias = {k: v.replace(' ', '_') for k, v in self.alias.items()}

    def get_topic(self, *args) -> str:
        args = [self.alias.get(a, a) for a in args]
        topic = self.base_topic + '/' + '/'.join(args)
        return topic.replace('//', '/')


class Publish(ConfigContainer):
    qos: int = ConfigEntry(default=0, description='Default QoS when publishing values')
    retain: bool = ConfigEntry(default=False, description='Default retain flag when publishing values')


class Mqtt(ConfigContainer):
    connection = Connection()
    topics = Topics()
    publish = Publish()


def log_lvl_validator(v, msg=''):
    if v not in logging._nameToLevel:
        raise Invalid(msg or f'Level must be one of {", ".join(logging._nameToLevel)}')

    return logging._nameToLevel[v]


class Logging(ConfigContainer):
    level: str = ConfigEntry(default='INFO', description='Verbosity level for the logfile',
                             validator=log_lvl_validator)
    file: Path = ConfigEntry(default_factory=lambda: str(CONFIG._path.with_name('sml2mqtt.log')),
                             description='Path to logfile')


class DeviceConfig:
    def __init__(self, device: str, timeout: float, skip: List[str]):
        self.device: str = device
        self.timeout: float = timeout
        self.skip: List[str] = skip

    @classmethod
    def create(cls, _in: dict, msg=None):
        if not isinstance(_in, dict) or not _in:
            raise Invalid(msg or 'Device config must be a dict')

        device = _in['device']
        if not isinstance(device, str) or not device:
            raise Invalid(msg or 'device must be a valid string')

        timeout = _in['timeout']
        if not isinstance(timeout, (float, int)) or timeout < 1:
            raise Invalid(msg or f'timeout must be a valid int/float and >= 1 ({type(timeout)}, {timeout})')

        skip = _in.get('skip', [])
        for i, v in enumerate(skip):
            if not isinstance(v, str) or not v:
                raise Invalid(msg or f'Skipped value at {i} must be a valid string')

        return cls(device, timeout, skip)


def device_validator(_in, msg=None):
    if not isinstance(_in, list):
        raise Invalid(msg or f'Devices must be a list (is {type(_in)})')
    return [DeviceConfig.create(e) for e in _in]


class General(ConfigContainer):
    max_wait: int = ConfigEntry(
        default=120, key_name='max wait', validator=Range(min=2),
        description='Time in seconds sml2mqtt waits for a value change until the value gets republished'
    )


class SmlMqttConfig(ConfigFile):
    mqtt = Mqtt()
    log = Logging()
    general = General()
    devices: List[DeviceConfig] = ConfigEntry(
        default_factory=lambda: [{'device': 'COM1', 'timeout': 3, 'skip': ['value ids that will', 'not be reported']},
                                 {'device': '/dev/ttyS0', 'timeout': 3}],
        description='Configuration of the sml devices', validator=device_validator)


CONFIG = SmlMqttConfig()



import logging
import sml2mqtt
import traceback
from asyncio import Future, create_task

from asyncio_mqtt import Client, MqttError, Will

from .config import CONFIG

log = logging.getLogger('sml.mqtt')

MQTT: Client = None
CONNECT: Future = None


async def _connect():
    global MQTT, CONNECT

    wait = 1

    while True:
        try:
            will_topic = CONFIG.mqtt.topics.get_topic(CONFIG.mqtt.topics.last_will)

            MQTT = Client(
                hostname=CONFIG.mqtt.connection.host,
                port=CONFIG.mqtt.connection.port,
                username=CONFIG.mqtt.connection.user if CONFIG.mqtt.connection.user else None,
                password=CONFIG.mqtt.connection.password if CONFIG.mqtt.connection.password else None,
                will=Will(will_topic, payload='OFFLINE')
            )

            # We don't publish anything if we just analyze the data from the reader
            if sml2mqtt._args.ARGS.analyze:
                return None

            await MQTT.connect()

            # signal that we are online
            await publish(will_topic, 'ONLINE')
            wait = 1
            break

        except MqttError as e:
            log.error(e)
            wait = min(180, wait * 2)
        except Exception:
            for line in traceback.format_exc().splitlines():
                log.error(line)
            return None

    CONNECT = None


async def connect():
    global CONNECT
    if CONNECT is None:
        CONNECT = create_task(_connect())


async def publish(topic, value):
    global CONNECT

    if not MQTT._client.is_connected():
        # try to connect
        if CONNECT is None:
            CONNECT = create_task(_connect())
        return None

    # publish message
    try:
        await MQTT.publish(topic, value, qos=CONFIG.mqtt.publish.qos, retain=CONFIG.mqtt.publish.retain)
    except MqttError as e:
        log.error(e)
